@@ -1,5 +1,6 @@
 import click
 
+from copy import deepcopy
 from typing import List, Union
 
 from astropy import wcs as pywcs # type: ignore
@@ -10,6 +11,12 @@ import healpy # type: ignore
 
 import numpy as np # type: ignore
 
+import logging
+
+logging.basicConfig()
+
+logger = logging.getLogger()
+
 debug = False
 
 def mosaic_list(in_fs: List[Union[str, fits.HDUList]], 
@@ -17,7 +24,8 @@ def mosaic_list(in_fs: List[Union[str, fits.HDUList]],
                 pixels="healpix", 
                 healpix_nsides=512, 
                 mock=False):
-    # print("input: %s output: %s"%( ", ".join(map(str, in_fs)), out_fn))
+
+    logger.debug("input: %s output: %s", ", ".join(map(str, in_fs)), out_fn)
 
     if pixels == "first":
         m = FITsMosaic(mock) # type: Mosaic
@@ -68,7 +76,7 @@ class Mosaic:
 
     def parse_in(self, in_f):
         if isinstance(in_f, str):
-            #print("parsing", in_f)
+            logger.debug("parsing %s", in_f)
             fn = in_f
             f = fits.open(in_f)
 
@@ -76,13 +84,13 @@ class Mosaic:
             f = in_f
             try:
                 fn = in_f.filename()
-                #print("parsing hdulist", fn)
+                logger.debug("parsing hdulist %s", fn)
             except Exception as e:
                 fn = "unknown"
-                #print("parsing hdulist with no detectable filename?")
+                logger.debug("parsing hdulist with no detectable filename?")
 
         else:
-            raise RuntimeError("unknown input " +repr(f))
+            raise RuntimeError("unknown input " + repr(in_f))
 
 
         img = None
@@ -90,11 +98,11 @@ class Mosaic:
         for n, img_parser in img_parsers.items():
             try:
                 img = img_parser(f)
-                #print("img parser succeeded", n)
+                logger.debug("img parser succeeded %s", n)
                 break
             except Exception as e:
-                print("img parser failed", n, e)
-                failures.append("img parser failed %s %s"%(n, e))
+                logger.info("img parser failed %s %s", n, e)
+                failures.append("img parser failed %s %s", n, e)
 
         if img is None:
             raise RuntimeError("all image parsers failed for %s: %s"%(f, ", ".join(failures)))
@@ -107,7 +115,7 @@ class Mosaic:
                 sj, si = map(int, img['wcs'].wcs_world2pix(83, 22, 0))
                 sjd, sid = map(int, img['wcs'].wcs_world2pix(83, 24, 0))
 
-                print("mock crab at", si, sj, "and", sid, sjd)
+                logger.info("mock crab at %s %s and %s %s", si, sj, sid, sjd)
 
                 img['flux'][
                         si-4:si+4,
@@ -118,24 +126,26 @@ class Mosaic:
                         sjd-4:sjd+4,
                         ] = 60
             except Exception as e:
-                print("unable to add mock crab", e)
+                logger.info("unable to add mock crab %s", e)
 
 
             img['flux'][
                     (int(img['flux'].shape[0]/2)-10):(int(img['flux'].shape[0]/2)+10),
                     (int(img['flux'].shape[1]/2)-10):(int(img['flux'].shape[1]/2)+10),
                     ] = 10
+
             img['flux'][
                     (int(img['flux'].shape[0]/2)-3):(int(img['flux'].shape[0]/2)+3),
                     (int(img['flux'].shape[1]/2)):(int(img['flux'].shape[1]/2)+30),
                     ] = 20
+
             fits.ImageHDU(img['flux'], header=img['wcs'].to_header()).writeto("mock-%s.fits"%fn.replace("/","_"), overwrite=True)
 
         return img
 
     @staticmethod
     def m_s(a, _s_a, b, _s_b):
-        c=dict()
+        c = dict()
 
         _m = ~np.isnan(_s_b(a['flux']))
 
@@ -159,25 +169,37 @@ class Mosaic:
     def m(a, _s_a, b, _s_b):
 
         def update_keywords():
+            # defined according to
+            # https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/ofwg_recomm/r11.html
+            # https://heasarc.gsfc.nasa.gov/docs/fcg/common_dict.html
 
-            additive_kw = ['ONTIME', 'EXPOSURE', 'TELAPSE']
+            additive_kw = ['ONTIME', 'EXPOSURE']
             incremental_kw_min = ['TSTART', 'DATE-OBS', 'TFIRST', 'DATE', 'MJD-OBS']
             incremental_kw_max = ['TSTOP', 'DATE-END', 'TLAST', 'MJD-END']
             average_kw = ['DEADC']
 
             keyword_list = [(additive_kw, lambda x, y: x + y),
                             (average_kw, lambda x, y: (x + y) / 2),
-                            (incremental_kw_min, lambda x, y: (x if x <= y else y)),
-                            (incremental_kw_max, lambda x, y: (x if x >= y else y))]
+                            (incremental_kw_min, lambda x, y: min(x, y)),
+                            (incremental_kw_max, lambda x, y: max(x, y))]
 
             keywords = dict()
 
-            for kk in keyword_list:
-                for keyword in kk[0]:
+            for kws, keyword_aggregator in keyword_list:
+                for keyword in kws:
                     try:
-                        keywords.update({keyword: kk[1](a['header'][keyword], b['header'][keyword])})
-                    except:
-                        Warning('Keyword ' + keyword + ' not found')
+                        keywords.update({
+                            keyword: keyword_aggregator(a['header'][keyword], b['header'][keyword])
+                        })
+                        logger.debug('%s (%s, %s) => %s', keyword, 
+                                                    str(a['header'][keyword]), 
+                                                    str(b['header'][keyword]),
+                                                    keywords[keyword])
+                    except KeyError as e:
+                        logger.warning(f'Keyword %s not found: %s', keyword, e)
+
+            if 'TSTART' in keywords and 'TSTOP' in keywords:
+                keywords['TELAPSE'] = keywords['TSTART'] - keywords['TSTOP']
 
             return keywords
 
@@ -187,17 +209,16 @@ class Mosaic:
         
         c = dict()
 
-        #print("m: _m shape", _m.shape)
-        #print("m: _m_a shape", _m_a.shape)
-        #print("m: _m_b shape", _m_b.shape)
+        logger.debug("m: _m shape %s", _m.shape)
+        logger.debug("m: _m_a shape %s", _m_a.shape)
+        logger.debug("m: _m_b shape %s", _m_b.shape)
 
         for k in 'flux', 'var', 'ex', 'n':
             c[k] = _s_a(a[k])
             c[k][~_m_a & _m_b] = _s_b(b[k])[~_m_a & _m_b]
 
-
-        #print("max exposure was", np.nanmax(a['ex']))
-        #print("max exposure to add", np.nanmax(b['ex']))
+        logger.debug("max exposure was %s", np.nanmax(a['ex']))
+        logger.debug("max exposure to add %s", np.nanmax(b['ex']))
 
         s_a = lambda x:_s_a(x)[_m]
         s_b = lambda x:_s_b(x)[_m]
@@ -210,7 +231,7 @@ class Mosaic:
         c['n'][_m] = s_a(a['n']) + s_b(b['n'])
         c['keywords'] = update_keywords()
 
-        #print("max exposure become", np.nanmax(c['ex']))
+        logger.debug("max exposure become %s", np.nanmax(c['ex']))
 
         return c
 
@@ -259,10 +280,10 @@ class FITsMosaic(Mosaic):
         img = self.parse_in(in_f)
 
         if self.mosaic is None: 
-            print("first mosaic")
+            logger.info("first mosaic")
             self.mosaic = img
         else:
-            print("adding mosaic")
+            logger.info("adding mosaic")
             m_j, m_i = np.meshgrid(
                     np.arange(self.mosaic['flux'].shape[1]),
                     np.arange(self.mosaic['flux'].shape[0]),
@@ -273,8 +294,8 @@ class FITsMosaic(Mosaic):
             i = i.astype(int)
             j = j.astype(int)
 
-            #print("MOSAIC:", self.mosaic['wcs'])
-            #print("IMG:", img['wcs'])
+            logger.debug("MOSAIC: %s", self.mosaic['wcs'])
+            logger.debug("IMG: %s", img['wcs'])
 
 
             self.mosaic.update(self.m(
@@ -283,6 +304,13 @@ class FITsMosaic(Mosaic):
                         img,
                         lambda x:pickornan(x, i, j),
                     ))
+            #we need to update the keywords because update is done in a static method
+            #that does not know about the keyword dictionary
+            if 'keywords' in self.mosaic.keys():
+                logger.debug('Updating keywords')
+                for kk, vv in self.mosaic['keywords'].items():
+                    self.mosaic['header'][kk] = vv
+
 
     def to_hdu_list(self):
         self.mosaic['sig'] = self.mosaic['flux'] / self.mosaic['var'] ** 0.5
@@ -291,7 +319,7 @@ class FITsMosaic(Mosaic):
 
         for k in self.mosaic:
 
-            #print(k)
+            logger.debug(k)
 
             if k == 'wcs' or k == 'keywords' or k == 'header':
                 continue
@@ -307,8 +335,9 @@ class FITsMosaic(Mosaic):
             h['IMATYPE'] = h['EXTNAME']
             if 'keywords' in self.mosaic.keys():
                 for key, value in self.mosaic['keywords'].items():
-                    #print('Update keyword %s = '%(key), value)
+                    logger.debug('Update keyword %s = %s', key, value)
                     h[key] = value
+                h['TELAPSE'] = (h['TLAST'] - h['TFIRST']) * 24 * 3600
             e = fits.ImageHDU(self.mosaic[k], header=h)
 
             el.append(e)
@@ -320,6 +349,7 @@ class FITsMosaic(Mosaic):
         self.to_hdu_list().writeto(fn, overwrite=True)
 
 
+
 class HealpixMosaic(Mosaic):
     def __init__(self, nsides=512, mock=False):
         self.nsides = nsides
@@ -378,14 +408,14 @@ class HealpixMosaic(Mosaic):
 
         ra, dec = self.radec_grid()
 
-        #print("max exposure", np.nanmax(self.mosaic['ex']))
+        logger.debug("max exposure %s", np.nanmax(self.mosaic['ex']))
 
         m_best = self.mosaic['ex'] > (np.nanmax(self.mosaic['ex'])/2.)
         c_ra = np.nanmean(ra[m_best])
         c_dec = np.nanmean(dec[m_best])
         c_rad_deg = 60
 
-        #print("center ra, dec", c_ra, c_dec)
+        logger.debug("center ra, dec %s %s", c_ra, c_dec)
 
         w = pywcs.WCS(naxis=2)
         ni, nj = 800, 800
@@ -397,6 +427,8 @@ class HealpixMosaic(Mosaic):
         w.wcs.cunit = 'deg', 'deg'
         w.wcs.cdelt = np.array([-pxsize, pxsize])
         w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
+
+        #TODO: is this needed
         #w.wcs.set_pv([(2, 1, 45.0)])
 
         m_j, m_i = np.meshgrid(
@@ -408,8 +440,8 @@ class HealpixMosaic(Mosaic):
 
         px = healpy.ang2pix(self.nsides, m_ra, m_dec, lonlat=True)
 
-
         self.mosaic['sig'] = self.mosaic['flux'] / self.mosaic['var']**0.5
+
         el = []
         for k in self.mosaic:
             if k == 'wcs' or k == 'keywords' or k == 'header':
@@ -423,11 +455,13 @@ class HealpixMosaic(Mosaic):
                         ex='EXPOSURE',
                         n='NIMAGE',
                     )[k]
+
             h['IMATYPE'] = h['EXTNAME']
             if 'keywords' in  self.mosaic.keys():
                 for key, value in self.mosaic['keywords'].items():
-                    #print('Update keyword %s = '%(key), value)
+                    logger.debug('Update keyword %s = %s',key, value)
                     h[key] = value
+
             mp = np.zeros((ni, nj))
             mp[:,:] = np.nan 
             mp[m_i, m_j] = self.mosaic[k][px]
@@ -435,6 +469,7 @@ class HealpixMosaic(Mosaic):
             e = fits.ImageHDU(mp, header=h)
 
             el.append(e)
+
         return fits.HDUList([fits.PrimaryHDU()] + el)
 
     def writeto_reproj(self, fn):
